"""
Python RMSD tool
"""

import numpy as np

from spyrmsd import molecule, rmsd


def coords_from_molecule(mol: molecule.Molecule, center: bool = False) -> np.ndarray:
    """
    Atomic coordinates from molecule.

    Parameters
    ----------
    mol: molecule.Molecule
        Molecule
    center: bool
        Center flag

    Returns
    -------
    np.ndarray
        Atomic coordinates (possibly centred)

    Notes
    -----
    Atomic coordinates are centred according to the center of geometry, not the center
    of mass.
    """

    if center:
        coords = mol.coordinates - mol.center_of_geometry()
    else:
        coords = mol.coordinates

    return coords


def rmsdwrapper(
    mol1, mol2, symmetry=True, center=False, minimize=False, strip=False
) -> float:
    """
    Compute RMSD between two molecule.

    Parameters
    ----------
    mol1: molecule.Molecule
        Molecule 1
    mol2: molecule.Molecule
        Molecule 2
    symmetry: bool, optional
        Symmetry-corrected RMSD (using graph isomorphism)
    center: bool, optional
        Center molecules at origin
    minimize: bool, optional
        Minimised RMSD (using the quaternion polynomial method)
    strip: bool, optional
        Strip hydrogen atoms

    Returns
    -------
    float
        RMSD
    """

    if strip:
        mol1.strip()  # Does nothing if already stripped
        mol2.strip()

    if minimize:
        center = True

    c1 = coords_from_molecule(mol1, center)
    c2 = coords_from_molecule(mol2, center)

    if c1.shape != c2.shape:
        # TODO: Create specific exception
        raise ValueError("Molecules have different sizes.")

    RMSD = np.inf

    if minimize and symmetry:
        RMSD = rmsd.rmsd_qcp_isomorphic(
            c1, c2, mol1.adjacency_matrix, mol2.adjacency_matrix
        )
    elif minimize and not symmetry:
        RMSD = rmsd.rmsd_qcp(c1, c2, mol1.atomicnums, mol2.atomicnums)
    elif not minimize and symmetry:
        RMSD = rmsd.rmsd_isomorphic(
            c1, c2, mol1.adjacency_matrix, mol2.adjacency_matrix
        )
    elif not minimize and not symmetry:
        RMSD = rmsd.rmsd_standard(c1, c2, mol1.atomicnums, mol2.atomicnums)

    return RMSD


if __name__ == "__main__":

    from pyrmsd import io

    import argparse as ap
    import os

    parser = ap.ArgumentParser(description="Python RMSD tool.")

    parser.add_argument("reference", type=str, help="Reference file")
    parser.add_argument("molecules", type=str, nargs="+", help="Input file(s)")
    parser.add_argument("-m", "--minimize", action="store_true", help="Minimize (fit)")
    parser.add_argument(
        "-c", "--center", action="store_true", help="Center molecules at origin"
    )
    parser.add_argument("-s", "--strip", action="store_true", help="Strip H atoms")
    parser.add_argument(
        "-n", "--nosymm", action="store_false", help="No graph isomorphism"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")

    args = parser.parse_args()

    output: str = ""

    obref = io.load(args.reference)
    ref = io.to_molecule(obref, adjacency=True)

    if args.verbose:
        refname = os.path.basename(args.reference)

    # Loop over input files
    for molfile in args.molecules:

        # Load all molecule within file
        obmols = io.loadall(molfile)
        mols = [io.to_molecule(obmol, adjacency=True) for obmol in obmols]

        if args.verbose:
            molname = os.path.basename(molfile)

            output = f"{refname}:{molname} "

        # Loop over molecules within file
        for idx, mol in enumerate(mols):

            r = rmsdwrapper(
                ref, mol, args.nosymm, args.center, args.minimize, args.strip
            )

            print(f"{output}{r:.5f}")



import copy
import sys

import pytest

from spyrmsd import molecule, spyrmsd
from tests import molecules


def test_spyrmsd_imported():
    assert "spyrmsd" in sys.modules


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_rmsdwrapper_molsize(mol: molecule.Molecule) -> None:

    m = copy.deepcopy(mol)
    ms = copy.deepcopy(mol)

    ms.strip()

    with pytest.raises(ValueError):
        spyrmsd.rmsdwrapper(m, ms, symmetry=False)


# Results obtained with MDAnalysis
#   trp0 = mda.Universe("trp0.pdb")
#   c0 = trp0.coord -trp0.select_atoms("protein").center_of_geometry()
#   for i in [1,2,3,4,5]:
#       trp = mda.Universe(f"trp{i}.pdb")
#       rmsd_dummy = mda.analysis.rms.rmsd(trp.coord, trp0.coord)
#       c = trp.coord - trp.select_atoms("protein").center_of_geometry()
#       _, rmsd_min = align.rotation_matrix(tc, tc0)
#       print(rmsd_dummy, rmsd_min)
@pytest.mark.parametrize(
    "i, rmsd_dummy, rmsd_min",
    [
        (1, 4.812480551076202, 1.6578281551053196),
        (2, 6.772045449820714, 1.7175638492348284),
        (3, 9.344911262612964, 1.5946081072641485),
        (4, 9.772939589989000, 2.1234944939308220),
        (5, 8.901837608843241, 2.4894805175766606),
    ],
)
def test_rmsdwrapper_qcp_protein(i: int, rmsd_dummy: float, rmsd_min: float):

    mol0 = copy.deepcopy(molecules.trp[0])
    mol = copy.deepcopy(molecules.trp[i])

    assert spyrmsd.rmsdwrapper(mol0, mol, symmetry=False) == pytest.approx(rmsd_dummy)
    assert spyrmsd.rmsdwrapper(
        mol0, mol, symmetry=False, minimize=True
    ) == pytest.approx(rmsd_min)


# Results obtained with OpenBabel
@pytest.mark.parametrize(
    "index, RMSD",
    [
        (1, 0.592256),
        (2, 2.11545),
        (3, 2.29824),
        (4, 9.45773),
        (5, 1.35005),
        (6, 9.44356),
        (7, 9.59758),
        (8, 9.55076),
        (9, 2.44067),
        (10, 9.6171),
    ],
)
def test_rmsdwrapper_isomorphic(index: int, RMSD: float) -> None:

    molc = copy.deepcopy(molecules.docking_1cbr[0])
    mol = copy.deepcopy(molecules.docking_1cbr[index])

    assert spyrmsd.rmsdwrapper(mol, molc, strip=True) == pytest.approx(RMSD, abs=1e-5)


# Results obtained with OpenBabel
@pytest.mark.parametrize(
    "index, RMSD",
    [
        (1, 0.476858),
        (2, 1.68089),
        (3, 1.50267),
        (4, 1.90623),
        (5, 1.01324),
        (6, 1.31716),
        (7, 1.11312),
        (8, 1.06044),
        (9, 0.965387),
        (10, 1.37842),
    ],
)
def test_rmsdwrapper_qcp_isomorphic(index: int, RMSD: float) -> None:

    molc = copy.deepcopy(molecules.docking_1cbr[0])
    mol = copy.deepcopy(molecules.docking_1cbr[index])

    assert spyrmsd.rmsdwrapper(mol, molc, minimize=True, strip=True) == pytest.approx(
        RMSD, abs=1e-5
    )



from typing import Any, Dict, List

import networkx as nx
import numpy as np
import qcelemental as qcel

# TODO: Move elsewhere?
connectivity_tolerance: float = 0.4


def graph_from_adjacency_matrix(adjacency_matrix: np.ndarray) -> nx.Graph:
    """
    Graph from adjacency matrix.

    Parameters
    ----------
    adjacency_matrix: np.ndarray
        Adjacency matrix

    Returns
    -------
    nx.Graph
        NetworkX graph
    """

    return nx.convert_matrix.from_numpy_array(adjacency_matrix)


def adjacency_matrix_from_atomic_coordinates(
    atomicnums: np.ndarray, coordinates: np.ndarray
) -> np.ndarray:
    """
    Compute adjacency matrix from atomic coordinates.

    Parameters
    ----------
    atomicnums: numpy.ndarray
        Atomic numbers
    coordinates: numpy.ndarray
        Atomic coordinates

    Returns
    -------
    numpy.ndarray
        Adjacency matrix

    Notes
    -----

    This function is based on an automatic bond perception algorithm: two
    atoms are considered to be bonded when their distance is smaller than
    the sum of their covalent radii plus a tolerance value. [3]_

    .. warning::
        The automatic bond perceptron rule implemented in this functions
        is very simple and only depends on atomic coordinates. Use
        with care!

    .. [3] E. C. Meng and R. A. Lewis, *Determination of molecular topology and atomic
       hybridization states from heavy atom coordinates*, J. Comp. Chem. **12**, 891-898
       (1991).
    """

    n = len(atomicnums)

    assert coordinates.shape == (n, 3)

    A = np.zeros((n, n))

    for i in range(n):
        r_i = qcel.covalentradii.get(atomicnums[i], units="angstrom")

        for j in range(i + 1, n):
            r_j = qcel.covalentradii.get(atomicnums[i], units="angstrom")

            distance = np.sqrt(np.sum((coordinates[i] - coordinates[j]) ** 2))

            if distance < (r_i + r_j + connectivity_tolerance):
                A[i, j] = A[j, i] = 1

    return A


def match_graphs(G1: nx.Graph, G2: nx.Graph) -> List[Dict[Any, Any]]:
    """
    Compute RMSD using the quaternion polynomial method.

    Parameters
    ----------
    G1: networkx.Graph
        Graph 1
    G2: networkx.Graph
        Graph 2

    Returns
    -------
    List[Dict[Any, Any]]
        All possible mappings between nodes of graph 1 and graph 2 (isomorphisms)

    Raises
    ------
    ValueError
        If the graphs `G1` and `G2` are not isomorphic
    """

    GM = nx.algorithms.isomorphism.GraphMatcher(G1, G2)

    # Check if graphs are actually isomorphic
    if not GM.is_isomorphic():
        # TODO: Create a new exception
        raise ValueError(f"Graphs {G1} and {G2} are not isomorphic.")

    # Extract all isomorphisms in a list
    return [isomorphism for isomorphism in GM.isomorphisms_iter()]



import numpy as np

from spyrmsd import graph, hungarian, qcp, utils


def rmsd_standard(
    coords1: np.ndarray,
    coords2: np.ndarray,
    atomicn1: np.ndarray,
    atomicn2: np.ndarray,
    center: bool = False,
) -> float:
    """
    Compute dummy (naïve) RMSD.

    Parameters
    ----------
    mol1: molecule.Molecule
        Molecule 1
    mol2: molecule.Molecule
        Molecule 2
    center: bool
        Flag for centering the molecules at the origin

    Returns
    -------
    float
        Dummy (naïve) RMSD

    Notes
    -----
    The dummy (naïve) RMSD distribution assumes a one-to-one mapping of the atoms in
    the order they are stored.

    The equality between all the atomic numbers is checked for sanity.
    """

    assert np.all(atomicn1 == atomicn2)
    assert coords1.shape == coords2.shape

    n = coords1.shape[0]

    # Center coordinates if required
    c1 = utils.center(coords1) if center else coords1
    c2 = utils.center(coords2) if center else coords2

    return np.sqrt(np.sum((c1 - c2) ** 2) / n)


def rmsd_qcp(
    coords1: np.ndarray, coords2: np.ndarray, atomicn1: np.ndarray, atomicn2: np.ndarray
) -> float:
    """
    Compute minimum RMSD using the Quaternion Characteristic Polynomial method.

    Parameters
    ----------
    coords1: np.ndarray
        Coordinate of molecule 1
    coords2: np.ndarray
        Coordinates of molecule 2
    atomicn1: np.ndarray
        Atomic numbers for molecule 1
    atomicn2: np.ndarray
        Atomic numbers for molecule 2

    Returns
    -------
    float
        Minimum RMSD (after superimposition)

    Notes
    -----
    The molecules are always centred at the origin according to the center of geometry
    and superimposed in order to minimize the RMSD. [1]_

    .. [1] D. L. Theobald, *Rapid calculation of RMSDs using a quaternion-based
       characteristic polynomial*, Acta Crys. A **61**, 478-480 (2005).
    """

    assert np.all(atomicn1 == atomicn2)

    # Center coordinates if required
    c1 = utils.center(coords1)
    c2 = utils.center(coords2)

    return qcp.qcp_rmsd(c1, c2)


def rmsd_hungarian(
    coords1: np.ndarray,
    coords2: np.ndarray,
    atomicn1: np.ndarray,
    atomicn2: np.ndarray,
    center=False,
):
    """
    Compute minimum RMSD using the Hungarian method.

    Parameters
    ----------
    coords1: np.ndarray
        Coordinate of molecule 1
    coords2: np.ndarray
        Coordinates of molecule 2
    atomicn1: np.ndarray
        Atomic numbers for molecule 1
    atomicn2: np.ndarray
        Atomic numbers for molecule 2

    Returns
    -------
    float
        Minimum RMSD (after assignment)

    Notes
    -----
    The Hungarian algorithm is used to solve the linear assignment problem, which is
    a minimum weight matching of the molecular graphs (bipartite). [2]_

    The linear assignment problem is solved for every element separately.

    .. [2] W. J. Allen and R. C. Rizzo, *Implementation of the Hungarian Algorithm to
        Account for Ligand Symmetry and Similarity in Structure-Based Design*,
        J. Chem. Inf. Model. **54**, 518-529 (2014)
    """

    assert atomicn1.shape == atomicn2.shape
    assert coords1.shape == coords2.shape

    # Center coordinates if required
    c1 = utils.center(coords1) if center else coords1
    c2 = utils.center(coords2) if center else coords2

    return hungarian.hungarian_rmsd(c1, c2, atomicn1, atomicn2)


def rmsd_isomorphic(
    coords1: np.ndarray,
    coords2: np.ndarray,
    am1: np.ndarray,
    am2: np.ndarray,
    center=False,
) -> float:
    """
    Compute minimum RMSD using graph isomorphism.

    Parameters
    ----------
    coords1: np.ndarray
        Coordinate of molecule 1
    coords2: np.ndarray
        Coordinates of molecule 2
    am1: np.ndarray
        Adjacency matrix for molecule 1
    am2: np.ndarray
        Adjacency matrix for molecule 2
    center: boolean
        Centering flag

    Returns
    -------
    float
        Minimum RMSD (after graph matching)
    """

    assert coords1.shape == coords2.shape

    n = coords1.shape[0]

    # Center coordinates if required
    c1 = utils.center(coords1) if center else coords1
    c2 = utils.center(coords2) if center else coords2

    # Convert molecules to graphs
    G1 = graph.graph_from_adjacency_matrix(am1)
    G2 = graph.graph_from_adjacency_matrix(am2)

    # Get all the possible graph isomorphisms
    isomorphisms = graph.match_graphs(G1, G2)

    # Minimum squared displacement
    min_sd = np.inf

    # Loop over all graph isomorphisms to find the lowest RMSD
    for isomorphism in isomorphisms:

        # Use the isomorphism to shuffle coordinates around (from original order)
        c1i = c1[list(isomorphism.keys()), :]
        c2i = c2[list(isomorphism.values()), :]

        # Compute square displacement
        # Avoid dividing by n and an expensive sqrt() operation
        sd = np.sum((c1i - c2i) ** 2)

        if sd < min_sd:
            min_sd = sd

    # Return the actual RMSD
    return np.sqrt(min_sd / n)


def rmsd_qcp_isomorphic(
    coords1: np.ndarray, coords2: np.ndarray, am1: np.ndarray, am2: np.ndarray
) -> float:
    """
    Compute minimum RMSD using the Quaternion Characteristic Polynomial method on
    isomorphic graphs.

    Parameters
    ----------
    coords1: np.ndarray
        Coordinate of molecule 1
    coords2: np.ndarray
        Coordinates of molecule 2
    am1: np.ndarray
        Adjacency matrix for molecule 1
    am2: np.ndarray
        Adjacency matrix for molecule 2

    Returns
    -------
    float
        Minimum RMSD (after graph matching and superimposition)

    Notes
    -----
    This QCP method works in cases where the atoms in `mol1` and `mol2` are not in the
    exact same order. If the atoms in `mol1` and `mol2` are in the same order use
    `rmsd_qcp` (faster).
    """

    assert coords1.shape == coords2.shape

    # Center coordinates
    c1 = utils.center(coords1)
    c2 = utils.center(coords2)

    # Build graph from adjacency matrix
    G1 = graph.graph_from_adjacency_matrix(am1)
    G2 = graph.graph_from_adjacency_matrix(am2)

    # Get all the possible graph isomorphisms
    isomorphisms = graph.match_graphs(G1, G2)

    # Minimum squared displacement
    min_rmsd = np.inf

    # Loop over all graph isomorphisms to find the lowest RMSD
    for isomorphism in isomorphisms:

        # Use the isomorphism to shuffle coordinates around (from original order)
        c1i = c1[list(isomorphism.keys()), :]
        c2i = c2[list(isomorphism.values()), :]

        rmsd = qcp.qcp_rmsd(c1i, c2i)

        if rmsd < min_rmsd:
            min_rmsd = rmsd

    return min_rmsd



"""
spyrmsd
Python RMSD tool
"""
import sys

from setuptools import find_packages, setup

import versioneer

short_description = __doc__.split("\n")

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except Exception:
    long_description = "\n".join(short_description[2:])


setup(
    # Self-descriptive entries which should always be present
    name="spyrmsd",
    author="Rocco Meli",
    author_email="rocco.meli@biodtp.ox.ac.uk",
    description=short_description[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="MIT",
    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages(),
    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    include_package_data=True,
    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,
    # Additional entries you may want simply uncomment the lines you want and fill in the data
    url="https://pyrmsd.readthedocs.io",  # Website
    # install_requires=[],              # Required packages, pulls from pip if needed; do not use for Conda deployment
    # platforms=['Linux',
    #            'Mac OS-X',
    #            'Unix',
    #            'Windows'],            # Valid platforms your code works on, adjust to your flavor
    python_requires=">=3.6",  # Python version restrictions
    # Manual control if final package is compressible or not, set False to prevent the .egg from being made
    # zip_safe=False,
)



import copy

import numpy as np
import pytest

from spyrmsd import molecule, rmsd
from tests import molecules


def test_rmsd_dummy_benzene() -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    assert rmsd.rmsd_standard(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)


def test_rmsd_dummy_shifted_benzene() -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.translate(np.array([0, 0, 1]))

    assert rmsd.rmsd_standard(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, ref=j, nofit=True)
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 2.60065218), (1, 3, 9.94411523), (2, 3, 9.4091711)]
)
def test_rmsd_dummy_2viz(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    assert rmsd.rmsd_standard(
        moli.coordinates, molj.coordinates, moli.atomicnums, molj.atomicnums
    ) == pytest.approx(result)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, mask="!@H=", ref=j, ref_mask="!@H=", nofit=True)
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 2.65327362), (1, 3, 10.11099065), (2, 3, 9.57099612)]
)
def test_rmsd_dummy_2viz_stripped(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    moli.strip()
    molj.strip()

    assert rmsd.rmsd_standard(
        moli.coordinates, molj.coordinates, moli.atomicnums, molj.atomicnums
    ) == pytest.approx(result)


def test_rmsd_dummy_centred_benzene() -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.translate(np.array([0, 0, 1]))

    assert rmsd.rmsd_standard(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1)
    assert rmsd.rmsd_standard(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        center=True,
    ) == pytest.approx(0)


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_rmsd_qcp(mol: molecule.Molecule) -> None:

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    assert rmsd.rmsd_standard(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)
    assert rmsd.rmsd_qcp(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)

    for _ in range(10):
        mol2.translate(np.random.rand(3))
        mol2.rotate(np.random.rand(1), np.random.rand(3))

        assert (
            rmsd.rmsd_standard(
                mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
            )
            > 0.0
        )

        assert rmsd.rmsd_qcp(
            mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
        ) == pytest.approx(0)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, ref=j)
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 1.95277757), (1, 3, 3.11801105), (2, 3, 2.98609758)]
)
def test_rmsd_qcp_2viz(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    assert rmsd.rmsd_qcp(
        moli.coordinates, molj.coordinates, moli.atomicnums, molj.atomicnums
    ) == pytest.approx(result)


# Results obtained with PyTraj
#   pytraj.analysis.rmsd.rmsd(i, "!@H=", ref=j, ref_mask="!@H=")
@pytest.mark.parametrize(
    "i, j, result", [(1, 2, 1.98171656), (1, 3, 3.01799306), (2, 3, 2.82917355)]
)
def test_rmsd_qcp_2viz_stripped(i: int, j: int, result: float) -> None:

    moli = copy.deepcopy(molecules.docking_2viz[i])
    molj = copy.deepcopy(molecules.docking_2viz[j])

    # Strip hydrogen atoms
    moli.strip()
    molj.strip()

    assert rmsd.rmsd_qcp(
        moli.coordinates, molj.coordinates, moli.atomicnums, molj.atomicnums
    ) == pytest.approx(result)


# Results obtained with MDAnalysis
#   trp0 = mda.Universe("trp0.pdb")
#   c0 = trp0.coord -trp0.select_atoms("protein").center_of_geometry()
#   for i in [1,2,3,4,5]:
#       trp = mda.Universe(f"trp{i}.pdb")
#       rmsd_dummy = mda.analysis.rms.rmsd(trp.coord, trp0.coord)
#       c = trp.coord - trp.select_atoms("protein").center_of_geometry()
#       _, rmsd_min = align.rotation_matrix(tc, tc0)
#       print(rmsd_dummy, rmsd_min)
@pytest.mark.parametrize(
    "i, rmsd_dummy, rmsd_min",
    [
        (1, 4.812480551076202, 1.6578281551053196),
        (2, 6.772045449820714, 1.7175638492348284),
        (3, 9.344911262612964, 1.5946081072641485),
        (4, 9.772939589989000, 2.1234944939308220),
        (5, 8.901837608843241, 2.4894805175766606),
    ],
)
def test_rmsd_qcp_protein(i: int, rmsd_dummy: float, rmsd_min: float):

    mol0 = copy.deepcopy(molecules.trp[0])
    mol = copy.deepcopy(molecules.trp[i])

    assert rmsd.rmsd_standard(
        mol0.coordinates, mol.coordinates, mol0.atomicnums, mol.atomicnums
    ) == pytest.approx(rmsd_dummy)
    assert rmsd.rmsd_qcp(
        mol0.coordinates, mol.coordinates, mol0.atomicnums, mol.atomicnums
    ) == pytest.approx(rmsd_min)


@pytest.mark.parametrize(
    "angle, tol", [(60, 1e-4), (120, 1e-4), (180, 1e-4), (240, 1e-4), (300, 1e-4)]
)
def test_rmsd_hungarian_benzene_rotated(angle: float, tol: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    assert rmsd.rmsd_standard(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)
    assert rmsd.rmsd_hungarian(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)

    # Rotations different than 180 degrees introduce numerical errors (~1e-6)
    mol2.rotate(angle, [0, 0, 1], units="deg")

    assert (
        rmsd.rmsd_standard(
            mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
        )
        > 0
    )
    assert rmsd.rmsd_hungarian(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0, abs=tol)


@pytest.mark.parametrize(
    "angle, tol", [(60, 1e-4), (120, 1e-4), (180, 1e-4), (240, 1e-4), (300, 1e-4)]
)
def test_rmsd_hungarian_benzene_shifted_rotated(angle: float, tol: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.translate([0, 0, 1])

    assert rmsd.rmsd_standard(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1)
    assert rmsd.rmsd_hungarian(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1)

    # Rotations different than 180 degrees introduce numerical errors (~1e-11)
    mol2.rotate(angle, [0, 0, 1], units="deg")

    assert (
        rmsd.rmsd_standard(
            mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
        )
        > 1
    )
    assert rmsd.rmsd_hungarian(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(1, abs=tol)


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_rmsd_hungarian_centred(mol: molecule.Molecule) -> None:

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    mol2.translate(np.random.rand(3))

    assert (
        rmsd.rmsd_hungarian(
            mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
        )
        > 0
    )
    assert rmsd.rmsd_hungarian(
        mol1.coordinates,
        mol2.coordinates,
        mol1.atomicnums,
        mol2.atomicnums,
        center=True,
    ) == pytest.approx(0)


@pytest.mark.parametrize("mol", molecules.allmolecules)
def test_rmsd_isomorphic_centred(mol: molecule.Molecule) -> None:

    mol1 = copy.deepcopy(mol)
    mol2 = copy.deepcopy(mol)

    mol2.translate(np.random.rand(3))

    assert (
        rmsd.rmsd_isomorphic(
            mol1.coordinates,
            mol2.coordinates,
            mol1.adjacency_matrix,
            mol2.adjacency_matrix,
        )
        > 0
    )
    assert rmsd.rmsd_isomorphic(
        mol1.coordinates,
        mol2.coordinates,
        mol1.adjacency_matrix,
        mol2.adjacency_matrix,
        center=True,
    ) == pytest.approx(0)


@pytest.mark.parametrize("angle", [60, 120, 180, 240, 300, 360])
def test_rmsd_isomorphic_rotated_benzene(angle: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.rotate(angle, np.array([0, 0, 1]), units="deg")

    assert (
        rmsd.rmsd_standard(
            mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
        )
        > 0
    )
    assert rmsd.rmsd_qcp(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)
    assert rmsd.rmsd_hungarian(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0, abs=1e-4)
    assert rmsd.rmsd_isomorphic(
        mol1.coordinates, mol2.coordinates, mol1.adjacency_matrix, mol2.adjacency_matrix
    ) == pytest.approx(0, abs=1e-4)


@pytest.mark.parametrize("angle", [60, 120, 180, 240, 300, 360])
def test_rmsd_isomorphic_rotated_benzene_stripped(angle: float) -> None:

    mol1 = copy.deepcopy(molecules.benzene)
    mol2 = copy.deepcopy(molecules.benzene)

    mol2.rotate(angle, np.array([0, 0, 1]), units="deg")

    mol1.strip()
    mol2.strip()

    assert (
        rmsd.rmsd_standard(
            mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
        )
        > 0
    )
    assert rmsd.rmsd_qcp(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0)
    assert rmsd.rmsd_hungarian(
        mol1.coordinates, mol2.coordinates, mol1.atomicnums, mol2.atomicnums
    ) == pytest.approx(0, abs=1e-4)
    assert rmsd.rmsd_isomorphic(
        mol1.coordinates, mol2.coordinates, mol1.adjacency_matrix, mol2.adjacency_matrix
    ) == pytest.approx(0, abs=1e-4)


# Results obtained with OpenBabel
@pytest.mark.parametrize(
    "index, RMSD",
    [
        (1, 0.592256),
        (2, 2.11545),
        (3, 2.29824),
        (4, 9.45773),
        (5, 1.35005),
        (6, 9.44356),
        (7, 9.59758),
        (8, 9.55076),
        (9, 2.44067),
        (10, 9.6171),
    ],
)
def test_rmsd_isomorphic(index: int, RMSD: float) -> None:

    molc = copy.deepcopy(molecules.docking_1cbr[0])
    mol = copy.deepcopy(molecules.docking_1cbr[index])

    molc.strip()
    mol.strip()

    assert rmsd.rmsd_isomorphic(
        molc.coordinates, mol.coordinates, molc.adjacency_matrix, mol.adjacency_matrix
    ) == pytest.approx(RMSD, abs=1e-5)


# Results obtained with OpenBabel
@pytest.mark.parametrize(
    "index, RMSD",
    [
        (1, 0.476858),
        (2, 1.68089),
        (3, 1.50267),
        (4, 1.90623),
        (5, 1.01324),
        (6, 1.31716),
        (7, 1.11312),
        (8, 1.06044),
        (9, 0.965387),
        (10, 1.37842),
    ],
)
def test_rmsd_qcp_isomorphic(index: int, RMSD: float) -> None:

    molc = copy.deepcopy(molecules.docking_1cbr[0])
    mol = copy.deepcopy(molecules.docking_1cbr[index])

    molc.strip()
    mol.strip()

    assert rmsd.rmsd_qcp_isomorphic(
        molc.coordinates, mol.coordinates, molc.adjacency_matrix, mol.adjacency_matrix
    ) == pytest.approx(RMSD, abs=1e-5)
